package geniusweb.logconverter;

import static org.junit.Assert.assertEquals;

import java.io.IOException;
import java.math.BigDecimal;
import java.net.URISyntaxException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;

import org.junit.Before;
import org.junit.Test;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;

import geniusweb.actions.PartyId;
import geniusweb.inform.Agreements;
import geniusweb.issuevalue.Bid;
import geniusweb.issuevalue.DiscreteValue;
import geniusweb.issuevalue.Value;
import geniusweb.protocol.session.SessionResult;
import geniusweb.references.Parameters;
import geniusweb.references.PartyRef;
import geniusweb.references.PartyWithParameters;
import geniusweb.references.PartyWithProfile;
import geniusweb.references.ProfileRef;
import tudelft.utilities.junit.GeneralTests;

public class SessionResultWithUtilsTest
		extends GeneralTests<SessionResultWithUtils> {

	private final ObjectMapper jackson = new ObjectMapper();
	private final RuntimeException error = new RuntimeException("test");

	private final String ISSUE1 = "salary";
	private final String ISSUE2 = "fte";
	private String jsonstring = "{\"participants\":{\"party2\":{\"party\":{\"partyref\":\"party2\",\"parameters\":{}},\"profile\":\"file:src/test/resources/profile2.json\"},\"party1\":{\"party\":{\"partyref\":\"party1\",\"parameters\":{}},\"profile\":\"file:src/test/resources/profile1.json\"}},\"agreements\":{\"party2\":{\"issuevalues\":{\"salary\":\"3500\",\"fte\":\"0.8\"}},\"party1\":{\"issuevalues\":{\"salary\":\"3500\",\"fte\":\"0.8\"}}},\"penalties\":{\"party2\":0.002,\"party1\":0.001},\"error\":null,\"utilities\":{\"party2\":0.7,\"party1\":0.8,\"party3\":0.4}}";
	private Map<PartyId, Double> penalties = new HashMap<>();

	private PartyId PARTY1 = new PartyId("party1");
	private PartyId PARTY2 = new PartyId("party2");
	private PartyId PARTY3 = new PartyId("party3");

	private SessionResult result1, result1a, result2, result3;
	private SessionResultWithUtils reswithutil1, reswithutil1a, reswithutil2;

	@Before
	public void before() throws URISyntaxException, JsonProcessingException {
		penalties.put(PARTY1, 0.001d);
		penalties.put(PARTY2, 0.002d);

		String errorstring = "\"error\":{\"java.lang.RuntimeException\":"
				+ jackson.writeValueAsString(error) + "}";
		// System.out.println(errorstring);

		PartyWithParameters party1 = new PartyWithParameters(
				new PartyRef("party1"), new Parameters());
		PartyWithParameters party2 = new PartyWithParameters(
				new PartyRef("party2"), new Parameters());

		PartyWithProfile partyprofile1 = new PartyWithProfile(party1,
				new ProfileRef("file:src/test/resources/profile1.json"));
		PartyWithProfile partyprofile2 = new PartyWithProfile(party2,
				new ProfileRef("file:src/test/resources/profile2.json"));

		Map<String, Value> issuevalues1 = new HashMap<>();
		issuevalues1.put(ISSUE1, new DiscreteValue("3500"));
		issuevalues1.put(ISSUE2, new DiscreteValue("0.8"));
		Bid bid1 = new Bid(issuevalues1);
		Agreements agreement1 = new Agreements().with(new Agreements(bid1,
				new HashSet<>(Arrays.asList(PARTY1, PARTY2))));

		// different order but that shouldn't matter
		Map<String, Value> issuevalues2 = new HashMap<>();
		issuevalues2.put(ISSUE1, new DiscreteValue("3500"));
		issuevalues2.put(ISSUE2, new DiscreteValue("0.8"));
		Bid bid2 = new Bid(issuevalues2);
		Agreements agreement2 = new Agreements().with(new Agreements(bid2,
				new HashSet<>(Arrays.asList(PARTY1, PARTY3))));

		Map<PartyId, PartyWithProfile> partiesmap = new HashMap<>();
		partiesmap.put(PARTY1, partyprofile1);
		partiesmap.put(PARTY2, partyprofile2);

		Map<PartyId, PartyWithProfile> partiesmap2 = new HashMap<>();
		partiesmap2.put(PARTY1, partyprofile2);
		partiesmap2.put(PARTY3, partyprofile1);

		result1 = new SessionResult(partiesmap, agreement1, penalties, null);
		result2 = new SessionResult(partiesmap, agreement2, penalties, null);
		result3 = new SessionResult(partiesmap, new Agreements(), penalties,
				null);

		HashMap<PartyId, BigDecimal> utils1 = new HashMap<>();
		utils1.put(PARTY1, new BigDecimal("0.8"));
		utils1.put(PARTY2, new BigDecimal("0.7"));
		utils1.put(PARTY3, new BigDecimal("0.4"));
		HashMap<PartyId, BigDecimal> utils2 = new HashMap<>();
		utils2.put(PARTY1, new BigDecimal("0.8"));
		utils2.put(PARTY2, new BigDecimal("0.9"));
		utils2.put(PARTY3, new BigDecimal("0.4"));

		reswithutil1 = new SessionResultWithUtils(partiesmap, agreement1,
				penalties, null, utils1);
		reswithutil1a = new SessionResultWithUtils(partiesmap, agreement1,
				penalties, null, utils1);
		reswithutil2 = new SessionResultWithUtils(partiesmap, agreement1,
				penalties, null, utils2);

	}

	@Override
	public List<List<SessionResultWithUtils>> getGeneralTestData() {
		return Arrays.asList(Arrays.asList(reswithutil1, reswithutil1a),
				Arrays.asList(reswithutil2));
	}

	@Override
	public List<String> getGeneralTestStrings() {
		return Arrays.asList(
				"SessionResultWithUtils.*party2=0.7, party1=0.8, party3=0.4.*",
				"SessionResultWithUtils.*party2=0.9, party1=0.8, party3=0.4.*");
	}

	@Test
	public void serializeTest() throws JsonProcessingException {
		System.out.println(jackson.writeValueAsString(reswithutil1));
		assertEquals(jsonstring, jackson.writeValueAsString(reswithutil1));
	}

	@Test
	public void deserializeTest() throws IOException {
		SessionResultWithUtils act = jackson.readValue(jsonstring,
				SessionResultWithUtils.class);
		assertEquals(reswithutil1, act);
	}

	@Test
	public void createTest() {
		TestReporter reporter = new TestReporter();
		SessionResultWithUtils res = SessionResultWithUtils.create(result1,
				reporter);
		System.out.println(res);
		assertEquals(null, reporter.getError());
		assertEquals(new BigDecimal("0.5590"), // 0.56 - 0.001
				res.getUtilities().get(PARTY1));
		assertEquals(new BigDecimal("0.448"), // 0.45 - 0.002
				res.getUtilities().get(PARTY2));

	}

	@Test
	public void createNoAgreementTest() {
		TestReporter reporter = new TestReporter();
		SessionResultWithUtils res = SessionResultWithUtils.create(result3,
				reporter);
		System.out.println(res);
		assertEquals(null, reporter.getError());
		assertEquals(new BigDecimal("0.5600"), // "salary": "3500","fte": "0.8"
				res.getUtilities().get(PARTY1));
		assertEquals(new BigDecimal("0.50"), // "salary": "3000", "fte": "0.8"
				res.getUtilities().get(PARTY2));

	}

	@Test
	public void blaTest() {
		// and if there is no resbid, is util=0?

	}

}
