package geniusweb.profile.utilityspace;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.io.IOException;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.junit.Before;
import org.junit.Test;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;

import geniusweb.issuevalue.Bid;
import geniusweb.issuevalue.DiscreteValue;
import geniusweb.issuevalue.DiscreteValueSet;
import geniusweb.issuevalue.Domain;
import geniusweb.issuevalue.Value;
import geniusweb.issuevalue.ValueSet;
import geniusweb.profile.Profile;
import tudelft.utilities.immutablelist.AbstractImmutableList;
import tudelft.utilities.immutablelist.ImmutableList;
import tudelft.utilities.immutablelist.Outer;

public class SumOfGroupsUtilitySpaceTest {
	private final ObjectMapper jackson = new ObjectMapper();
	private SumOfGroupsUtilitySpace pa1;
	private final String iss1 = "issue1", iss2 = "issue2", iss3 = "issue3";
	private static BigDecimal N01 = new BigDecimal("0.1"),
			N02 = new BigDecimal("0.2"), N03 = new BigDecimal("0.3"),
			N05 = new BigDecimal("0.5"), N07 = new BigDecimal("0.7"),
			N08 = new BigDecimal("0.8"), N09 = new BigDecimal("0.9");

	private final DiscreteValue lo = new DiscreteValue("low");
	private final DiscreteValue hi = new DiscreteValue("high");
	private final Bid resBid = new Bid(iss1, hi);

	private Domain domain;
	private PartsUtilities putils12, putils3;
	private String serialized = "{\"SumOfGroupsUtilitySpace\":{\"domain\":{\"name\":\"domain\",\"issuesValues\":{\"issue3\":{\"values\":[\"low\",\"high\"]},\"issue2\":{\"values\":[\"low\",\"high\"]},\"issue1\":{\"values\":[\"low\",\"high\"]}}},\"name\":\"testspace\",\"partUtilities\":{\"1and2\":{\"PartsUtilities\":{\"issues\":[\"issue1\",\"issue2\"],\"utilslist\":[{\"values\":[\"low\",\"low\"],\"util\":0.16},{\"values\":[\"high\",\"high\"],\"util\":0.72},{\"values\":[\"low\",\"high\"],\"util\":0.16},{\"values\":[\"high\",\"low\"],\"util\":0.16}]}},\"3\":{\"PartsUtilities\":{\"issues\":[\"issue3\"],\"utilslist\":[{\"values\":[\"low\"],\"util\":0.16},{\"values\":[\"high\"],\"util\":0.14}]}}},\"reservationbid\":{\"issuevalues\":{\"issue1\":\"high\"}}}}";

	@Before
	public void before() {
		Map<String, ValueSet> issues = new HashMap<>();
		issues.put(iss1, new DiscreteValueSet(lo, hi));
		issues.put(iss2, new DiscreteValueSet(lo, hi));
		issues.put(iss3, new DiscreteValueSet(lo, hi));
		domain = new Domain("domain", issues);

		Map<ProductOfValue, BigDecimal> utils12 = new HashMap<>();
		// util12 has weight N08
		utils12.put(new ProductOfValue(Arrays.asList(lo, lo)),
				N02.multiply(N08));
		utils12.put(new ProductOfValue(Arrays.asList(lo, hi)),
				N02.multiply(N08));
		utils12.put(new ProductOfValue(Arrays.asList(hi, lo)),
				N02.multiply(N08));
		utils12.put(new ProductOfValue(Arrays.asList(hi, hi)),
				N09.multiply(N08));
		putils12 = new PartsUtilities(Arrays.asList(iss1, iss2), utils12);

		// utils3 has weight N02.
		Map<ProductOfValue, BigDecimal> utils3 = new HashMap<>();
		utils3.put(new ProductOfValue(Arrays.asList(lo)), N08.multiply(N02));
		utils3.put(new ProductOfValue(Arrays.asList(hi)), N07.multiply(N02));
		putils3 = new PartsUtilities(Arrays.asList(iss3), utils3);

	}

	@Test(expected = IllegalArgumentException.class)
	public void smokeTestIncompleteProfile() {
		HashMap<String, PartsUtilities> partutils = new HashMap<>();
		partutils.put("1and2", putils12);
		new SumOfGroupsUtilitySpace(domain, "testspace", partutils, resBid);

	}

	@Test // non-normalized utils should be fine in this case
	public void smokeTestWeightsNotNormalized() {
		HashMap<String, PartsUtilities> partutils = new HashMap<>();
		partutils.put("1and2", putils12);
		partutils.put("3", putils3);
		new SumOfGroupsUtilitySpace(domain, "testspace", partutils, resBid);

	}

	@Test(expected = IllegalArgumentException.class)
	public void smokeTestExtraIssue() {
		HashMap<String, PartsUtilities> partutils = new HashMap<>();
		partutils.put("1and2", putils12);
		partutils.put("3", putils3);
		partutils.put("4", putils3);
		new SumOfGroupsUtilitySpace(domain, "testspace", partutils, resBid);

	}

	@Test
	public void smokeTest() {
		makeBasic();
	}

	@Test
	public void testName() {
		SumOfGroupsUtilitySpace paus = makeBasic();
		assertEquals("testspace", paus.getName());
	}

	@Test
	public void testDomain() {
		SumOfGroupsUtilitySpace paus = makeBasic();
		assertEquals(domain, paus.getDomain());
	}

	@Test
	public void testGetResBid() {
		SumOfGroupsUtilitySpace paus = makeBasic();
		assertEquals(resBid, paus.getReservationBid());
	}

	@Test
	public void testGetUtilityEmptyBid() {
		SumOfGroupsUtilitySpace paus = makeBasic();
		assertTrue(BigDecimal.ZERO.compareTo(
				paus.getUtility(new Bid(Collections.emptyMap()))) == 0);
	}

	@Test
	public void testGetUtility() {
		SumOfGroupsUtilitySpace paus = makeBasic();
		Map<String, Value> map = new HashMap<>();
		map.put(iss1, lo);
		map.put(iss2, lo); // 1,2 combined has util 0.2
		map.put(iss3, hi); // 3 has util 0.7
		// weights are set to 0.8, 0.2. Gives total util 0.3
		Bid bid = new Bid(map);
		assertTrue(N03.compareTo(paus.getUtility(bid)) == 0);
	}

	/**
	 * Demonstrates how to convert {@link LinearAdditiveUtilitySpace} to a
	 * {@link SumOfGroupsUtilitySpace}
	 * 
	 */
	@Test
	public void testConvertLas() throws IOException {
		String profile = new String(
				Files.readAllBytes(Paths.get("src/test/resources/party1.json")),
				StandardCharsets.UTF_8);
		LinearAdditiveUtilitySpace las = jackson.readValue(profile,
				LinearAdditiveUtilitySpace.class);
		SumOfGroupsUtilitySpace pas = new SumOfGroupsUtilitySpace(las);
		checkUtilitiesAllEqual(las, pas);
		System.out.println("Generated partial profile:\n" + pas);
	}

	/**
	 * Demonstrates how to convert {@link LinearAdditiveUtilitySpace} to a
	 * {@link SumOfGroupsUtilitySpace} and group some issues
	 * 
	 */
	@Test
	public void testConvertAndGroupParty() throws IOException {
		String profile = new String(
				Files.readAllBytes(Paths.get("src/test/resources/party1.json")),
				StandardCharsets.UTF_8);
		LinearAdditiveUtilitySpace las = jackson.readValue(profile,
				LinearAdditiveUtilitySpace.class);
		SumOfGroupsUtilitySpace pas = new SumOfGroupsUtilitySpace(las);
		pas = pas.group(Arrays.asList("Food", "Drinks"), "food&drinks");
		checkUtilitiesAllEqual(las, pas);
		System.out.println("Re-grouped party profile:\n");
		System.out.println(jackson.writeValueAsString(pas));

	}

	@Test
	public void testGroupe2e() {
		Map<ProductOfValue, BigDecimal> utils1 = new HashMap<>();
		utils1.put(new ProductOfValue(Arrays.asList(lo)), N02.multiply(N02));
		utils1.put(new ProductOfValue(Arrays.asList(hi)), N03.multiply(N02));
		Map<ProductOfValue, BigDecimal> utils2 = new HashMap<>();
		utils2.put(new ProductOfValue(Arrays.asList(lo)), N07.multiply(N03));
		utils2.put(new ProductOfValue(Arrays.asList(hi)), N02.multiply(N03));
		Map<ProductOfValue, BigDecimal> utils3 = new HashMap<>();
		utils3.put(new ProductOfValue(Arrays.asList(lo)), N05.multiply(N05));
		utils3.put(new ProductOfValue(Arrays.asList(hi)), N08.multiply(N05));
		HashMap<String, PartsUtilities> partutils = new HashMap<>();
		partutils.put(iss1, new PartsUtilities(Arrays.asList(iss1), utils1));
		partutils.put(iss2, new PartsUtilities(Arrays.asList(iss2), utils2));
		partutils.put(iss3, new PartsUtilities(Arrays.asList(iss3), utils3));

		Map<String, ValueSet> issvalues = new HashMap<>();
		issvalues.put(iss1, new DiscreteValueSet(lo, hi));
		issvalues.put(iss2, new DiscreteValueSet(lo, hi));
		issvalues.put(iss3, new DiscreteValueSet(lo, hi));
		Domain domain1 = new Domain("test", issvalues);

		SumOfGroupsUtilitySpace pas = new SumOfGroupsUtilitySpace(domain1,
				"testspace", partutils, resBid);

		SumOfGroupsUtilitySpace pasgrouped12 = pas
				.group(Arrays.asList(iss1, iss2), "iss12");

		checkEqualUtilities(pas, pasgrouped12);

	}

	@Test
	public void testSerialize() throws JsonProcessingException {
		SumOfGroupsUtilitySpace paus = makeBasic();
		ObjectMapper jackson = new ObjectMapper();
		String string = jackson.writeValueAsString(paus);
		System.out.println(string);
		assertEquals(serialized, string);

	}

	@Test
	public void testDeserialize() throws IOException {
		ObjectMapper jackson = new ObjectMapper();
		assertEquals(makeBasic(), jackson.readValue(serialized, Profile.class));

	}

	/************************ PRIVATE FUNCTIONS *******************/

	private void checkEqualUtilities(UtilitySpace space1, UtilitySpace space2) {
		for (Bid bid : new MyAllBidsList(space1.getDomain())) {
			assertEquals(space1.getUtility(bid), space2.getUtility(bid));
		}
	}

	private SumOfGroupsUtilitySpace makeBasic() {
		HashMap<String, PartsUtilities> partutils = new HashMap<>();
		partutils.put("1and2", putils12);
		partutils.put("3", putils3);
		return new SumOfGroupsUtilitySpace(domain, "testspace", partutils,
				resBid);

	}

	private void checkUtilitiesAllEqual(UtilitySpace space1,
			UtilitySpace space2) {
		for (Bid bid : new MyAllBidsList(space1.getDomain())) {
			assertTrue(space1.getUtility(bid)
					.compareTo(space2.getUtility(bid)) == 0);
		}
	}

}

/**
 * Copy of AllBidsList. We can not depend on BidSpace because that would cause a
 * cyclic dependency.
 */
class MyAllBidsList extends AbstractImmutableList<Bid> {

	private final List<String> issues;
	private final Outer<? extends Value> allValuePermutations;

	/**
	 * A list containing all bids in the space. This is an ImmutableList so it
	 * can contain all bids without pre-computing them.
	 * 
	 */
	public MyAllBidsList(Domain domain) {
		if (domain == null)
			throw new IllegalArgumentException("domain=null");
		issues = new ArrayList<>(domain.getIssues());

		List<ImmutableList<Value>> values = issues.stream()
				.map(issue -> domain.getValues(issue))
				.collect(Collectors.toList());

		allValuePermutations = new Outer<Value>(values);
	}

	@Override
	public Bid get(BigInteger index) {
		ImmutableList<? extends Value> nextValues = allValuePermutations
				.get(index);

		Map<String, Value> issueValues = new HashMap<>();
		for (int n = 0; n < issues.size(); n++) {
			issueValues.put(issues.get(n),
					nextValues.get(BigInteger.valueOf(n)));
		}
		return new Bid(issueValues);
	}

	@Override
	public BigInteger size() {
		return allValuePermutations.size();
	}

}
